/* Target-dependent code for OSF/1 on Alpha.
   Copyright 2002 Free Software Foundation, Inc.

   This file is part of GDB.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.  */

#include "defs.h"
#include "frame.h"
#include "gdbcore.h"
#include "value.h"

#include "alpha-tdep.h"

/* Under OSF/1, the __sigtramp routine is frameless and has a frame
   size of zero, but we are able to backtrace through it.  */
static CORE_ADDR
alpha_osf1_skip_sigtramp_frame (struct frame_info *frame, CORE_ADDR pc)
{
  char *name;

  find_pc_partial_function (pc, &name, (CORE_ADDR *) NULL, (CORE_ADDR *) NULL);
  if (PC_IN_SIGTRAMP (pc, name))
    return frame->frame;
  return 0;
}

static int
alpha_osf1_pc_in_sigtramp (CORE_ADDR pc, char *func_name)
{
  return (func_name != NULL && STREQ ("__sigtramp", func_name));
}

static CORE_ADDR
alpha_osf1_sigcontext_addr (struct frame_info *frame)
{
  return (read_memory_integer (frame->next ? frame->next->frame
					   : frame->frame, 8));
}

/* Read an instruction from memory at PC, looking through breakpoints.
   Imported from GDB-6.0:alpha-tdep.c.  */

static unsigned int
alpha_read_insn (CORE_ADDR pc)
{
  char buf[4];
  int status;

  status = read_memory_nobpt (pc, buf, 4);
  if (status)
    memory_error (status, pc);
  return extract_unsigned_integer (buf, 4);
}

static int
alpha_osf1_at_function_start (CORE_ADDR pc, CORE_ADDR func_start)
{
  /* A function prologue typically starts with 2 instructions setting
     up the GP.  But the linker sometimes changes the jump address of
     subroutine calls because it has determined that these 2 instructions
     do not make any difference in the execution of the program.  So it
     has virtually optimized them out by changing the jump address
     inside the function call to branch just after these 2 instructions.

     As a consequence, the "effective" start address of a given function
     can be either the first instruction of this function, or the third
     instruction if the first two instructions are setting up the GP.
     
     It is important to recognize these two cases during a "next" in
     order to properly detect function calls that land into a prologueless
     function which first two instructions have been shunted by the linker.  */

  if (pc == func_start)
    return 1;

  if (pc == func_start + 8)
    {
      unsigned int inst;

      /* ldah $gp,n($t12) */
      inst = alpha_read_insn (func_start);
      if ((inst & 0xffff0000) != 0x27bb0000)
        return 0;

      /* lda $gp,n($gp) */
      inst = alpha_read_insn (func_start + 4);
      if ((inst & 0xffff0000) != 0x23bd0000)
        return 0;

      return 1;
    }

  return 0;
}

static void
alpha_osf1_init_abi (struct gdbarch_info info,
                     struct gdbarch *gdbarch)
{
  struct gdbarch_tdep *tdep = gdbarch_tdep (gdbarch);

  set_gdbarch_pc_in_sigtramp (gdbarch, alpha_osf1_pc_in_sigtramp);
  /* The next/step support via procfs on OSF1 is broken when running
     on multi-processor machines. We need to use software single stepping
     instead.  */
  set_gdbarch_software_single_step (gdbarch, alpha_software_single_step);
  set_gdbarch_at_function_start (gdbarch, alpha_osf1_at_function_start);

  tdep->skip_sigtramp_frame = alpha_osf1_skip_sigtramp_frame;
  tdep->sigcontext_addr = alpha_osf1_sigcontext_addr;

  tdep->jb_pc = 2;
  tdep->jb_elt_size = 8;
}

void
_initialize_alpha_osf1_tdep (void)
{
  gdbarch_register_osabi (bfd_arch_alpha, GDB_OSABI_OSF1, alpha_osf1_init_abi);
}
